{$A+,B-,D+,E+,F-,G+,I-,L+,N+,O-,R-,S-,V-,X+}
{$M 16384,0,655360}
program ObjectsDemo;
uses Dos;               { fr Registers, Intr}

const Scale = 14;        { hher setzen bei schnelleren Rechnern }
      ObjCount = 100;    { evtl. bei langsamen Rechnern verkleinern }

type
     TPoint = record    { ein Punkt }
       X, Y: integer;
      end;

     TPixel = record    { ein Pixel }
       X, Y: integer;   { Koordinaten }
       Color: byte;     { und Farbe }
     end;

     PPixelArray = ^TPixelArray;
     TPixelArray = array[0..0] of TPixel; { maximalgroes Pixelarray ($R-)!}

{ ------ Objekt-Typen ------- }
     PBouncer = ^TBouncer;      { Zeiger auf.., fr dynamische Deklaration }
     TBouncer = object          { Das Ausgangsobjekt, Vorfahre aller anderen }
       Size: integer;           { Gre ist allen gemeinsam (und quadratisch) }
       Origin, Delta: TPoint;   { auch Position und Geschwindigkeit }
       constructor Init;        { mu gebaut werden }
       destructor Done; virtual;
       procedure Move; virtual; { "wei" wie es sich zu bewegen hat }
       procedure Draw(Show: boolean); virtual; { abstrakt, fr Move }
     end;

     PColorBouncer = ^TColorBouncer;
     TColorBouncer = object(TBouncer)   { erster Nachkomme, immernoch abstrakt }
       Color: byte;                     { hat zustzlich Farbe }
       constructor Init;                { ..die initialisiert werden mu }
       procedure CalcDraw(var x1, y1, x2, y2: integer); { neue Koordinaten }
     end;

     PCircle = ^TCircle;
     TCircle = object(TColorBouncer)    { was anschauliches: ein Kreis }
       Radius: integer;                 { hat einen Radius }
       constructor Init;                { der berechnet werden mchte }
       procedure Draw(Show: boolean); virtual; { kann sich zeichnen }
     end;

     PSquare = ^TSquare;
     TSquare = object(TColorBouncer)    { ein Quadrat, erbt alles restliche }
       procedure Draw(Show: boolean); virtual; { mu sich nur zeichnen knnen }
     end;

     PCross = ^TCross;
     TCross = object(TColorBouncer)     { ein Kreuz }
       procedure Draw(Show: boolean); virtual; { "wei" sich zu zeichnen }
     end;

     PXign = ^TXign;
     TXign = object(TColorBouncer)      { ein X }
       procedure Draw(Show: boolean); virtual; { wird wieder anders gemalt }
     end;

     PCloud = ^TCloud;
     TCloud = object(TBouncer)  { eine Pixelwolke hat keine einheitliche Farbe }
       Cloud: PPixelArray;      { aber ein Pixelarray mit entsp. Information }
       constructor Init;                { mu deshalb neu gebaut werden }
       destructor Done; virtual;        { und zum schlu auch demontiert }
       procedure Draw(Show: boolean); virtual; { und natrlich geamlt }
     end;

{ ------- Dienstroutinen ------- }

procedure PutCircle(X, Y, Radius, Color: integer); far; external;
{$L CIRC.OBJ } { fr besonders schnelle Kreise }

function IsVga: boolean;        { haben wir VGA }
var Regs: Registers;
begin
  with Regs do
  begin
    AX:= $1A00;                 { Display Combination abfragen }
    Intr($10, Regs);
    IsVga:= (AL = $1A);         { gltiger Aufruf }
  end;
end;

procedure SetVGAMode(Mode: byte);       { Setzt einen VGA-Modus }
var Regs: Registers;
begin
  with Regs do
  begin
    AX:= Mode;                  { Mode-Nr. in AL, AH:= 0 }
    Intr($10, Regs);
  end;
  Port[$3CE]:= 3;               { Index fr Data Rotate Reg. }
  Port[$3CF]:= 0;               { berschreiben einstellen (NormalPut) }
end;

procedure SetColor(Color: Byte); { Setzt Farbe fr sptere Schreibaktionen }
begin
  Port[$3CE]:= 0;               { Index fr Set/Reset Register }
  Port[$3CF]:= Color;           { Farbnummer laden }
end;

procedure EnterPixel;           { Leitet Pixel-malen ein }
begin
  Port[$3CE]:= 1;               { Index fr Enable Set/Reset Reg. }
  Port[$3CF]:= $0F;             { Set/Reset als Datenquelle }
end;

procedure DrawPixel(x,y: integer); { malt ein Pixel, Farbe mit SetColor }
var ScrOfs: Word;
    BitMask: Byte;
begin
  ScrOfs:= y * 80 + x shr 3;    { Bildspeicher-Offsetadresse }
  BitMask:= $80 shr (x and 7);  { Bitmaske berechnen }
  Port[$3CE]:= 8;               { Index fr Bit Mask Register }
  Port[$3CF]:= BitMask;         { Bitmaske laden }
  Inc(Mem[$A000:ScrOfs]);       { Lese/Schreibzyklus erzeugen! }
end;

procedure ExitPixel;            { beendet Pixel-malen (Register-Default) }
begin
  Port[$3CF]:= $FF;             { Bit Mask Reg. mit Default laden }
  Port[$3CE]:= 1;               { Index fr Enable Set/Reset Reg. }
  Port[$3CF]:= 0;               { mit Default laden }
end;

procedure HLine(x1,x2,y: Word);         { eine schnelle horizontale Linie }
var ScrOfs, HLength, x: Word;
    LMask, RMask: Byte;                 { Bitmasken fr den Rand }
begin
{ --- Vorbereitung --- }
  if x1 > x2 then Exit;                 { Wenn x1 > x2, sofort beenden }
  ScrOfs:= y * 80 + x1 div 8;           { Bildspeicher-Offsetadresse }
  HLength:= x2 div 8 - x1 div 8;        { Lnge der Linie in Bytes - 1 }
  LMask:= $FF shr (x1 and 7);           { linke Randmaske berechnen }
  RMask:= $FF shl (7-(x2 and 7));       { rechte Randmaske berechnen }
  if HLength = 0 then                   { wenn nur ein Byte dann linke }
    LMask:= LMask and RMask;            { und rechte Maske verknpfen }
  Port[$3CE]:= 1;                       { Index Enable Set/Reset Reg. }
  Port[$3CF]:= $0F;                     { Set/Reset als Datenquelle }
{ --- Linie zeichnen --- }
  Port[$3CE]:= 8;                       { Index fr Bit Mask Register }
  Port[$3CF]:= LMask;                   { Randmaske fr erstes Byte }
  Inc( Mem[$A000:ScrOfs] );             { Lese/Schreibzyklus erzeugen }
  if HLength > 0 then                   { Falls Lnge - 1 > 0: }
  begin
    Dec(HLength);                       { verbleibende Lnge }
    Inc(ScrOfs);                        { nchste Bildspeicheradresse }
    if HLength > 0 then                 { Falls Lnge - 1 > 0: }
    begin
      Port[$3CF]:= $FF;                 { Bitmaske FFh laden }
      FillChar(Mem[$A000:ScrOfs],       { Mittlerer Teil der Linie }
               HLength, $FF);
      inc(ScrOfs,HLength);              { Offsetadresse erhhen }
    end;
    Port[$3CF]:= RMask;                 { Randmaske fr letztes Byte }
    Inc( Mem[$A000:ScrOfs] );           { Lese/Schreibzyklus erzeugen }
  end;
{ --- Register aufrumen --- }
  Port[$3CF]:= $FF;                     { Default-Bitmaske }
  Port[$3CE]:= 1;                       { Index fr Enable Set/Reset }
  Port[$3CF]:= 0;                       { Enable Set/Reset-Default }
end;

procedure VLine(x,y1,y2: Word);         { eine schnelle senkrechte Linie }
var ScrOfs,y: Word;
    BitMask: Byte;
begin
  ScrOfs:= y1 * 80 + x div 8;           { Bildspeicher-Offsetadresse }
  BitMask:= $80 shr (x and 7);          { Bitmaske berechnen }
  Port[$3CE]:= 1;                       { Index fr Enable Set/Reset }
  Port[$3CF]:= $0F;                     { Set/Reset als Datenquelle }
  Port[$3CE]:= 8;                       { Index fr Bit Mask Register }
  Port[$3CF]:= BitMask;                 { Bitmaske laden }
  for y:= y1 to y2 do
  begin
    Inc( Mem[$A000:ScrOfs] );           { Lese/Schreibzyklus erzeugen }
    inc(ScrOfs, 80);                    { Offsetadr. eine Zeile weiter }
  end;
  Port[$3CE]:= 8;                       { Index fr Bit Mask Register }
  Port[$3CF]:= $FF;                     { Default-Bitmaske }
  Port[$3CE]:= 1;                       { Index fr Enable Set/Reset }
  Port[$3CF]:= 0;                       { Enable Set/Reset-Default }
end;

function GetRandColor: byte;            { eine zufllige aber helle Farbe }
begin
  GetRandColor:= Random(8) + 7;
end;

function KeyPressed: boolean;           { Taste gedrckt }
var
  ChainStart:  word absolute $40:$1A;   { Beginn des Pufferinhalts }
  ChainEnd:    word absolute $40:$1C;   { Beginn des freien Bereichs }
begin
  KeyPressed:= ChainStart <> ChainEnd;  { was dazwischen: erwartet Abholung }
  ChainStart:= ChainEnd;
end;

function MouseInstalled: boolean;       { Maus da ? }
var i33: pointer;
    Regs: Registers;
begin
  GetIntvec($33, i33);                  { Vektor holen }
  MouseInstalled:= false;
  if i33 <> nil then                    { Zeigt er wohin ? }
  with Regs do                          { Ok: Aufruf wagen }
  begin
    AX:= 0;                             { Funktion 0: Reset }
    intr($33, Regs);
    MouseInstalled:= ax = $FFFF;        { ax = -1 wenn treiber da }
  end;
end;

function MouseMoved: boolean;           { wurde die Maus bewegt? }
var Regs: Registers;
begin
  Regs.AX:= 11;                         { Funktion 11: relative Bewegung }
  Intr($33, Regs);
  MouseMoved:= Regs.CX or Regs.DX <> 0; { irgendwas geschehen? }
end;

{ ------- Objekte ------- }

{ TBouncer = object
    Size: integer;
    Origin, Delta: TPoint;}
constructor TBouncer.Init;              { Initialisiert den Ur-Vorfahren }
var HSize: integer;
begin
  Size:= Random(60) + 20;               { Gre zwischen 20 und 80 Pixel }
  HSize:= Size div 2;
  Origin.X:= Scale * 320 - HSize;       { Beginnt immer horizontal zentiert }
  Origin.Y:= Scale * 150 - HSize;       { und etwas oberhalb der Mitte }
  Delta.X:= Random(64) - integer(32);   { waagerechte Geschwindigkeit  32 }
  Delta.Y:= Random(74) - integer(24);   { senkrecht 50..98 }
end;

destructor TBouncer.Done;               { Zerstrt den Ur-Vorfahren }
begin
  { abstract }                          { nichts zu tun }
end;

procedure TBouncer.Move;                { Bewegt auch alle Nachkommen }
var New: TPoint;                        { neue Koordinate }
begin
  New.X:= Origin.X + Delta.X;           { neu:= alt + Geschwindigkeit }
  New.Y:= Origin.Y + Delta.Y;
  if (New.X >= Scale * (639 - Size)) or { Rechter.. }
     (New.X <= 0) then                  { ..oder linker Rand? }
  begin
    dec(New.X, Delta.X);                { dann nderung rckgngig machen }
    Delta.X:= -Delta.X;                 { Geschwindigkeit umkehren }
  end;
  if (New.Y >= Scale * (479 - Size)) or { das gleiche fr Oben und Unten }
     (New.Y <= 0) then
  begin
    dec(New.Y, Delta.Y);
    Delta.Y:= -Delta.Y;
  end;
  Draw(false);                  { jeweiliges Objekt lschen (ruft Nachkommen auf!) }
  Origin:= New;                 { neue Position eintragen }
  Draw(true);                   { und neu Zeichen }
  inc(Delta.Y);                 { Beschleunigung }
end;

procedure TBouncer.Draw(Show: boolean); { abstrakte Zeichenfunktion }
begin                                   { TBouncer dient nur als Vorfahre }
  { Abstract }
end;

{ TColorBouncer = object(TBouncer);
    Color: byte; }
constructor TColorBouncer.Init;
begin
{$IFDEF VER70}
  inherited Init;       { BP7 kennt "unbekannte" Vererbung }
{$ELSE}
  TBouncer.Init;        { in TP6 mu der Vorfahr explizit angegeben werden }
{$ENDIF}
  Color:= GetRandColor; { Farbe initialisieren }
end;

procedure TColorBouncer.CalcDraw(var x1, y1, x2, y2: integer); { neue Pos. }
begin
  with Origin do
  begin
    x1:= X div Scale;   { Aus Rechenposition tatschliche Berechnen }
    y1:= Y div Scale;
    x2:= x1 + Size;     { auch Gre }
    y2:= y1 + Size;
  end;
end;

{ TCircle = object(TColorBouncer) }
constructor TCircle.Init;       { das erste sichtbare Objekt }
begin
{$IFDEF VER70}
  inherited Init;
{$ELSE}
  TColorBouncer.Init;
{$ENDIF}
  Radius:= Size div 2;          { initialisiert nur den Radius.. }
end;

procedure TCircle.Draw(Show: boolean); { .. aber wei sich zu malen }
begin
  with Origin do
   if Show then        { zeigen? }
     PutCircle(X div Scale + Radius, Y div Scale + Radius, Radius, Color) {ja}
   else
     PutCircle(X div Scale + Radius, Y div Scale + Radius, Radius, 0); { nein }
end;

{ TSquare = object(TColorBouncer) }
procedure TSquare.Draw(Show: boolean); { ein Quadrat malen }
var x1, y1 ,x2, y2: integer;
begin
  if Show then
    SetColor(Color)             { Zeigen: Farbe setzen }
  else
    SetColor(0);                { sonst Schwarz }
  CalcDraw(x1, y1, x2, y2);     { Pixel-Ecken berechnen }
  HLine(x1, x2, y1);            { Horizontale Linien malen }
  HLine(x1, x2, y2);
  VLine(x1, y1, y2);            { und vertikale }
  VLine(x2, y1, y2);
end;

{ TCross = object(TColorBouncer) }
procedure TCross.Draw(Show: boolean);   { ein Kreuz }
var x1, y1 ,x2, y2: integer;
begin
  if Show then
    SetColor(Color)
  else                                  { Farbe setzen }
    SetColor(0);
  CalcDraw(x1, y1, x2, y2);             { Ecken berechnen }
  HLine(x1, x2, y1 + Size div 2);       { Linien durch die Mitte }
  VLine(x1 + Size div 2, y1, y2);
end;

{ TXign = object(TColorBouncer) }
procedure TXign.Draw(Show: boolean);    { Ein X }
var x1, y1 ,x2, y2: integer;
begin
  EnterPixel;                   { einzelne Pixel malen einleiten }
  if Show then
    SetColor(Color)
  else                          { Farbe setzen }
    SetColor(0);
  CalcDraw(x1, y1, x2, y2);     { Ecken holen }
  for x1:= x1 to x2 do          { von links nach rechts }
  begin
    DrawPixel(x1, y1);          { Pixel oben }
    DrawPixel(x1, y2);          { Pixel unten }
    inc(y1); dec(y2);           { und Abstand verkleinern }
  end;
  ExitPixel;                    { Pixel malen beenden }
end;

{ TCloud = object(TBouncer)
    Cloud: PPixelArray; }
constructor TCloud.Init;        { eine Punktwolke }
var i: integer;
    Rad, Ang, S2: real;
begin
{$IFDEF VER70}
  inherited Init;               { Vorfahre initialisieren }
{$ELSE}
  TBouncer.Init;
{$ENDIF}
  S2:= Size / 2;                                { kommt fter vor }
  GetMem(Cloud, Size * 4 * SizeOf(TPixel));     { Platz fr die Wolke mit.. }
  for i:= 0 to Size * 4 - 1 do                  { .. 4 * Size Pixel }
  with Cloud^[i] do
  begin
    Rad:= Random * S2;                  { ein zuflliger Mittenabstand }
    Ang:= Random * 2 * Pi;              { dito ein Winkel zwischen 0 & 360 }
    X:= round(Rad * sin(Ang) + S2);     { in Koordinaten konvertieren }
    Y:= round(Rad * cos(Ang) + S2);     { relativ zum linken, oberen Eck }
    Color:= GetRandColor;               { jeder Punkt ist anders gefrbt }
  end;
end;

destructor TCloud.Done;                 { Wolke zerstren }
begin
  FreeMem(Cloud, Size * 4 * SizeOf(TPixel));    { Speicher freigeben }
{$IFDEF VER70}
  inherited Done;                       { Vorfahre zerstren }
{$ELSE}
  TBouncer.Done;                        { dito fr Pascal 6 }
{$ENDIF}

end;

procedure TCloud.Draw(Show: boolean);   { Wolke malen }
var i, x1, y1: integer;
begin
  x1:= Origin.X div Scale;              { Ecke vorberechnen }
  y1:= Origin.Y div Scale;
  EnterPixel;                           { pixelmalen einleiten }
  if Show then                          { Zeigen? }
   for i:= 0 to Size * 4 - 1 do         { fr alle Punkte.. }
    with Cloud^[i] do
    begin
      SetColor(Color);                  { ..Farbe setzen.. }
      DrawPixel(x1 + X, y1 + Y);        { und malen }
    end
  else
  begin                                 { Lschen? }
    SetColor(0);                        { alle schwarz }
    for i:= 0 to Size * 4 - 1 do        { alle Punkte }
     with Cloud^[i] do
      DrawPixel(x1 + X, y1 + Y);        { schwrzen }
  end;
  ExitPixel;                            { pixelmalen beenden }
end;

{ ------- Variablen des Hauptprogramms ------- }

var Objects: array[0..ObjCount] of PBouncer; { ein Array mit Objekten }
    i: integer;                              { Zhler dafr }
    MouseOk: boolean;                        { merkt sich ob die Maus geht }

begin
  if not IsVga then             { gibt's VGA }
  begin
    writeln('Luft leider nur auf VGA-Karten...');
    writeln('Sorry...');
  end;
  MouseOk:= MouseInstalled;     { gibt's Maus? merken! }
  SetVGAMode($12);              { 640x480x16 einschalten }
  writeln('ObjectDemo - bewegt unterschiedliche Objekte (c) Erik Krause, Freiburg `95');
  writeln(' Beispielprogramm zum Artikel OBJEKTE aus DOS-Trend');
  writeln(' Ende durch Tastendruck oder Mausbewegung...');
  while Keypressed do;                  { Tastenpuffer leeren }
  if MouseOk then while MouseMoved do;  { Mausbewegungen abfangen }
  for i:= 0 to ObjCount-1 do            { jede Menge Objekte.. }
    case Random(5) of                   { ..per Zufall ausgewhlt }
      0: Objects[i]:= New(PCircle, Init);
      1: Objects[i]:= New(PSquare, Init);
      2: Objects[i]:= New(PCloud, Init);
      3: Objects[i]:= New(PCross, Init);
      4: Objects[i]:= New(PXign, Init);
    end; {case}
  repeat                        { die Hauptschleife tut nichts weiter.. }
    for i:= 1 to ObjCount-1 do  { ..als fr alle Objekte die Methode Move.. }
      Objects[i]^.Move;         { ..des gemeinsamen Vorfahren aufzurufen }
  until KeyPressed or           { bis eine Taste gedrckt }
        (MouseOk and MouseMoved); { oder die Maus bewegt wird }
  for i:= 1 to ObjCount-1 do      { alle Objekte.. }
    Dispose(Objects[i], Done);    { vom Heap entfernen }
  SetVGAMode(3);                  { Textmodus einstellen }
end.
